#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>

#include <analysis/Polynomial1D.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("Polynomial1D", "[analysis]")
{
  SECTION("Constructor")
  {
    Polynomial1D p({-1, 0, 0, 1});

    REQUIRE(p.coefficient(0) == -1);
    REQUIRE(p.coefficient(1) == 0);
    REQUIRE(p.coefficient(2) == 0);
    REQUIRE(p.coefficient(3) == 1);

    REQUIRE(p.degree() == 3);
  }

  SECTION("value")
  {
    Polynomial1D p({-1, -1.3, 2.7, 1});

    REQUIRE(p.degree() == 3);

    double x = 1.3;
    REQUIRE(p(x) == Catch::Approx(-1 - 1.3 * x + 2.7 * x * x + x * x * x));

    x = 0;
    REQUIRE(p(x) == Catch::Approx(-1 - 1.3 * x + 2.7 * x * x + x * x * x));

    x = 7;
    REQUIRE(p(x) == Catch::Approx(-1 - 1.3 * x + 2.7 * x * x + x * x * x));

    x = -3;
    REQUIRE(p(x) == Catch::Approx(-1 - 1.3 * x + 2.7 * x * x + x * x * x));
  }

  SECTION("opposite")
  {
    Polynomial1D p({-1, -1.3, 2.7, 1});
    Polynomial1D r = -p;

    REQUIRE(r.degree() == 3);

    REQUIRE(r.coefficient(0) == 1);
    REQUIRE(r.coefficient(1) == 1.3);
    REQUIRE(r.coefficient(2) == -2.7);
    REQUIRE(r.coefficient(3) == -1);
  }

  SECTION("sum")
  {
    Polynomial1D p({-1, 0, 0, 1});
    Polynomial1D q({-1, 2, 1.4, 2, -1});

    Polynomial1D r = p + q;
    REQUIRE(r.degree() == 4);

    REQUIRE(r.coefficient(0) == -2);
    REQUIRE(r.coefficient(1) == 2);
    REQUIRE(r.coefficient(2) == 1.4);
    REQUIRE(r.coefficient(3) == 3);
    REQUIRE(r.coefficient(4) == -1);
  }

  SECTION("difference")
  {
    Polynomial1D p({-1, 0, 0, 1});
    Polynomial1D q({-1, 2, 1.4, 2, -1});
    {
      Polynomial1D r = p - q;
      REQUIRE(r.degree() == 4);

      REQUIRE(r.coefficient(0) == 0);
      REQUIRE(r.coefficient(1) == -2);
      REQUIRE(r.coefficient(2) == -1.4);
      REQUIRE(r.coefficient(3) == -1);
      REQUIRE(r.coefficient(4) == 1);
    }

    {
      Polynomial1D s = q - p;
      REQUIRE(s.degree() == 4);

      REQUIRE(s.coefficient(0) == 0);
      REQUIRE(s.coefficient(1) == 2);
      REQUIRE(s.coefficient(2) == 1.4);
      REQUIRE(s.coefficient(3) == 1);
      REQUIRE(s.coefficient(4) == -1);
    }

    {
      Polynomial1D s = q - q;
      REQUIRE(s.degree() == 0);

      REQUIRE(s.coefficient(0) == 0);
    }
  }

  SECTION("product")
  {
    Polynomial1D p({-1, 1, 0, 2});
    Polynomial1D q({-1, 2, 1.4, 2, -1});

    {
      Polynomial1D s = 2 * p;
      REQUIRE(s.degree() == 3);

      REQUIRE(s.coefficient(0) == -2);
      REQUIRE(s.coefficient(1) == 2);
      REQUIRE(s.coefficient(2) == 0);
      REQUIRE(s.coefficient(3) == 4);
    }

    {
      Polynomial1D r = p * q;
      REQUIRE(r.degree() == 7);

      REQUIRE(r.coefficient(0) == Catch::Approx(1));
      REQUIRE(r.coefficient(1) == Catch::Approx(-3));
      REQUIRE(r.coefficient(2) == Catch::Approx(0.6));
      REQUIRE(r.coefficient(3) == Catch::Approx(-2.6));
      REQUIRE(r.coefficient(4) == Catch::Approx(7));
      REQUIRE(r.coefficient(5) == Catch::Approx(1.8));
      REQUIRE(r.coefficient(6) == Catch::Approx(4));
      REQUIRE(r.coefficient(7) == Catch::Approx(-2));
    }
  }

  SECTION("divide")
  {
    SECTION("exact")
    {
      Polynomial1D p({-1, 1, 0, 2});
      Polynomial1D q({1, -3, 0.6, -2.6, 7, 1.8, 4, -2});

      {
        Polynomial1D s = q / p;

        REQUIRE(s.degree() == 4);

        REQUIRE(s.coefficient(0) == -1);
        REQUIRE(s.coefficient(1) == 2);
        REQUIRE(s.coefficient(2) == 1.4);
        REQUIRE(s.coefficient(3) == 2);
        REQUIRE(s.coefficient(4) == -1);
      }
    }

    SECTION("with remainder")
    {
      Polynomial1D p({-1, 0, 1});
      Polynomial1D q({0, 1});

      {
        Polynomial1D s = p / q;

        REQUIRE(s.degree() == 1);

        REQUIRE(s.coefficient(0) == 0);
        REQUIRE(s.coefficient(1) == 1);
      }
    }
  }

  SECTION("mod")
  {
    SECTION("exact")
    {
      Polynomial1D p({-1, 1, 0, 2});
      Polynomial1D q({1, -3, 0.6, -2.6, 7, 1.8, 4, -2});

      {
        Polynomial1D s = q % p;

        REQUIRE(s.degree() == 0);
        REQUIRE(s.coefficient(0) == 0);
      }
    }

    SECTION("with remainder")
    {
      Polynomial1D p({-1, 0, 1});
      Polynomial1D q({0, 1});

      {
        Polynomial1D s = p % q;

        REQUIRE(s.degree() == 0);
        REQUIRE(s.coefficient(0) == -1);
      }
    }
  }

  SECTION("derive")
  {
    SECTION("constant")
    {
      Polynomial1D p(std::vector<double>{-1});
      Polynomial1D p_prime = derive(p);

      REQUIRE(p_prime.degree() == 0);
      REQUIRE(p_prime.coefficient(0) == 0);
    }

    SECTION("non constant")
    {
      Polynomial1D p({-1, 1, 0, 2});
      Polynomial1D p_prime = derive(p);

      REQUIRE(p_prime.degree() == 2);
      REQUIRE(p_prime.coefficient(0) == 1);
      REQUIRE(p_prime.coefficient(1) == 0);
      REQUIRE(p_prime.coefficient(2) == 6);
    }
  }

  SECTION("primitive")
  {
    SECTION("zero")
    {
      Polynomial1D p(std::vector<double>{0});
      Polynomial1D P = primitive(p);

      REQUIRE(P.degree() == 0);
      REQUIRE(P.coefficient(0) == 0);
    }

    SECTION("non zero")
    {
      Polynomial1D p({-1, 1, 0, 2});
      Polynomial1D P = primitive(p);

      REQUIRE(P.degree() == 4);
      REQUIRE(P.coefficient(0) == 0);
      REQUIRE(P.coefficient(1) == -1);
      REQUIRE(P.coefficient(2) == 0.5);
      REQUIRE(P.coefficient(3) == 0);
      REQUIRE(P.coefficient(4) == 0.5);
    }
  }

  SECTION("Output")
  {
    Polynomial1D p({-1, 0, 0, 1});

    std::ostringstream os;
    os << p;

    REQUIRE(os.str() == "-1 + 1*x^3");
  }

#ifndef NDEBUG
  SECTION("checking for bounds validation")
  {
    Polynomial1D p({-1, 0, 0, 1});
    REQUIRE_THROWS_AS(p.coefficient(4), AssertError);

    const Polynomial1D q({-1, 0, 0, 1});
    REQUIRE_THROWS_AS(q.coefficient(4), AssertError);

    REQUIRE_THROWS_AS(Polynomial1D{std::vector<double>{}}, AssertError);
  }
#endif   // NDEBUG
}
